# Octave Gain in dB
OCTAVE_GAIN_MIN = -20
OCTAVE_GAIN_MAX = 20
OCTAVE_GAIN_STEP = 0.5
# OPX+ waveform amplitude in V
OPX_AMPLITUDE_MIN = -0.5
OPX_AMPLITUDE_MAX = 0.5 - 2**-28
# OPX1000 MW-FEM waveform amplitude
OPX1000_MW_AMPLITUDE_MIN = -1.0
OPX1000_MW_AMPLITUDE_MAX = 1.0
# OPX1000 MW-FEM full scale power in dBm
OPX1000_MW_POWER_MIN = -11
OPX1000_MW_POWER_MAX = 16
OPX1000_MW_POWER_STEP = 3
# OPX1000 MW-FEM bands
OPX1000_MW_BANDS = {1: (50e6, 5.5e9), 2: (4.5e9, 7.5e9), 3: (6.5e9, 10.5e9)}
